import pytest

from pycantonese.util import _split_chars_with_alphanum


@pytest.mark.parametrize(
    "chars, expected",
    [
        ("廣東話", ("廣", "東", "話")),
        ("proposal", ("proposal",)),
        ("hap唔happy", ("hap", "唔", "happy")),
        ("high-tech唔好用", ("high-tech", "唔", "好", "用")),
        ("唔好成日開party", ("唔", "好", "成", "日", "開", "party")),
        ("死chur爛chur", ("死", "chur", "爛", "chur")),
    ],
)
def test_split_characters_with_english(chars, expected):
    assert _split_chars_with_alphanum(chars) == expected



import pycantonese


HKCANCOR = pycantonese.hkcancor()


def test_hkcancor_word_count():
    assert len(HKCANCOR.words()) == 153_654


def test_hkcancor_character_count():
    assert len(HKCANCOR.characters()) == 191_851



from functools import wraps
from string import ascii_letters, digits
import warnings


# What defines non-Cantonese-ness in lettered.json from rime-cantonese
_NOT_CANTONESE = ascii_letters + digits + "-"


def _split_chars_with_alphanum(chars):
    """
    Split Cantonese characters while respecting alphanumeric characters.

    :param chars: String of Cantonese chars, possibly with alphanumeric chars.

    :return: The split result that respects the alphanumeric characters.

    :rtype: tuple of str
    """
    if not chars:
        return tuple()
    if len(chars) == 1:
        return tuple(chars)
    result = []
    first = chars[0]
    for second in chars[1:]:
        if first[-1] in _NOT_CANTONESE and second in _NOT_CANTONESE:
            first += second
        else:
            result.append(first)
            first = second
    result.append(first)
    return tuple(result)


def _deprecate(what, use_instead, since, remove_from):
    """Create a decorator which throws a FutureWarning.

    FutureWarning is used instead of DeprecationWarning, because Python
    does not show DeprecationWarning by default.

    Parameters
    ----------
    what : str
        What to deprecate.
    use_instead : str
        Use this instead.
    since : str
        Version "x.y.z" since which the deprecation is in effect.
    remove_from : str
        Version "x.y.z" after which the deprecated functionality is removed.

    Returns
    -------
    A decorated function that throws a FutureWarning.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"'{what}' has been deprecated since PyCantonese v{since} and "
                f"will be removed from v{remove_from}. Please use "
                f"'{use_instead}' instead.",
                FutureWarning,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator



from functools import lru_cache

from wordseg import LongestStringMatching

from pycantonese.corpus import hkcancor
from pycantonese.data.rime_cantonese import CHARS_TO_JYUTPING, LETTERED
from pycantonese.util import _split_chars_with_alphanum


_MAX_WORD_LENGTH = 5

_ALLOWED_WORDS = None
_DISALLOWED_WORDS = None


class Segmenter(LongestStringMatching):
    """A customizable word segmentation model.

    .. versionadded:: 3.0.0
    """

    def __init__(
        self,
        *,
        max_word_length=_MAX_WORD_LENGTH,
        allow=None,
        disallow=None,
    ):
        """Initialize a Segmenter object.

        Parameters
        ----------
        max_word_length : int, optional
            Maximum word length this model allows.
        allow : iterable[str], optional
            Words to allow in word segmentation.
        disallow : iterable[str], optional
            Words to disallow in word segmentation.
        """
        super(Segmenter, self).__init__(max_word_length=max_word_length)

        # Train with HKCanCor data.
        self.fit(hkcancor().words(by_utterances=True))

        # Train with rime-cantonese data.
        self._words |= CHARS_TO_JYUTPING.keys()
        self._words |= LETTERED.keys()

        # Adjust with the allowed and disallowed words.
        self._words |= allow or set()
        self._words -= disallow or set()

        # Turn everything from strings to tuples due to alphanumeric chars.
        self._words = {_split_chars_with_alphanum(x) for x in self._words}

    def _predict_sent(self, sent_str):
        chars = _split_chars_with_alphanum(sent_str)
        segmented = super(Segmenter, self)._predict_sent(chars)
        # Turn the result back from tuples to strings.
        segmented = ["".join(x) for x in segmented]
        return segmented


@lru_cache(maxsize=1)
def _get_default_segmenter():
    return Segmenter()


def segment(unsegmented, cls=None):
    """Segment the unsegmented input.

    The word segmentation model is the longest string matching approach,
    trained by (i) the HKCanCor corpus included in this library and
    (ii) the rime-cantonese data.
    The segmented sentence does not contain words longer than five
    characters.

    .. versionadded:: 2.4.0

    .. versionchanged:: 3.0.0
        Added the keyword argument ``cls`` to allow a customized segmenter.

    Parameters
    ----------
    unsegmented : str
        Unsegmented input.
    cls: Segmenter, optional
        A custom `Segmenter` class object for setting the maximal
        word length (default = 5) and words to allow or disallow.
        If not provided, a default segmenter is used, with maximum word
        length = 5.

    Returns
    -------
    list[str]

    Examples
    --------
    >>> segment("廣東話容唔容易學？")  # "Is Cantonese easy to learn?"
    ['廣東話', '容', '唔', '容易', '學', '？']
    >>>
    >>> # Customizing the segmentation behavior.
    >>> from pycantonese.word_segmentation import Segmenter
    >>> segmenter = Segmenter(allow={"容唔容易"})
    >>> segment("廣東話容唔容易學？", cls=segmenter)
    ['廣東話', '容唔容易', '學', '？']
    """
    if not unsegmented:
        return []
    if cls is None:
        cls = _get_default_segmenter()
    if type(cls) != Segmenter:
        raise TypeError(f"`segmenter` must be a Segmenter object: {cls}")
    # Strip all whitespace.
    unsegmented = "".join(unsegmented.split())
    segmented = list(cls.predict([unsegmented]))[0]
    return segmented



import pytest

from pycantonese import pos_tag


def test_pos_tag():
    expected = [("我", "PRON")]
    assert pos_tag(["我"]) == expected
    assert pos_tag(["我"], tagset="universal") == expected


def test_pos_tag_hkcancor_tagset():
    assert pos_tag(["我"], tagset="hkcancor") == [("我", "R")]


def test_pos_tag_wrong_input_type():
    with pytest.raises(TypeError):
        pos_tag("我")


def test_pos_tag_unknown_tagset():
    with pytest.raises(ValueError):
        pos_tag(["我"], tagset="unknown tagset")



from pycantonese.corpus import Token
from pycantonese.tests.test_corpus import HKCANCOR


def test_find_verbs_in_hkcancor():
    all_verbs = HKCANCOR.search(pos="^V")
    assert len(all_verbs) == 29726
    assert all_verbs[:10] == [
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
        Token(word="有冇", pos="V1", jyutping="jau5mou5", mor=None, gra=None),
        Token(word="要", pos="VU", jyutping="jiu3", mor=None, gra=None),
        Token(word="有得", pos="VU", jyutping="jau5dak1", mor=None, gra=None),
        Token(word="冇得", pos="VU", jyutping="mou5dak1", mor=None, gra=None),
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="係", pos="V", jyutping="hai6", mor=None, gra=None),
        Token(word="係", pos="V", jyutping="hai6", mor=None, gra=None),
    ]



"""POS tagset mapping between HKCanCor and Universal Dependencies."""

from pycantonese.pos_tagging.punctuation_marks import _PUNCTUATION_MARKS


# The Python dictionary below maps the HKCanCor tagset to the Universal
# Dependencies (UD) 2.0 tagset.
#
# HKCanCor tagset: http://compling.hss.ntu.edu.sg/hkcancor/
# UD 2.0 tagset: https://universaldependencies.org/u/pos/index.html
#
# The HKCanCor paper describes 46 tags in its tagset, but the
# actual data (as included in this pycantonese library) has 112 tags,
# all of which are listed as keys in the Python dictionary below.
#
# For convenience, if HKCanCor has a brief part-of-speech tag description,
# the description appears as a comment together with the key in the
# dictionary below, e.g., the key "A" has the comment "HKCanCor: Adjective".

_MAP = {
    "!": "PUNCT",
    '"': "PUNCT",
    "#": "X",
    "'": "PUNCT",
    ",": "PUNCT",
    "-": "PUNCT",
    ".": "PUNCT",
    "...": "PUNCT",
    "?": "PUNCT",
    "A": "ADJ",  # HKCanCor: Adjective
    "AD": "ADV",  # HKCanCor: Adjective as Adverbial
    "AG": "ADJ",  # HKCanCor: Adjective Morpheme
    "AIRWAYS0": "PROPN",
    "AN": "NOUN",  # HKCanCor: Adjective with Nominal Function
    "AND": "PROPN",  # In one instance of "Chilli and Pepper"
    "B": "ADJ",  # HKCanCor: Non-predicate Adjective
    "BG": "ADJ",  # HKCanCor: Non-predicate Adjective Morpheme
    "BEAN0": "PROPN",  # In one instance of "Mr Bean"
    "C": "CCONJ",  # HKCanCor: Conjunction
    "CENTRE0": "NOUN",  # In one instance of "career centre"
    "CG": "CCONJ",
    "D": "ADV",  # HKCanCor: Adverb
    "D1": "ADV",  # Most instances are gwai2 "ghost".
    "DG": "ADV",  # HKCanCor: Adverb Morpheme
    "E": "INTJ",  # HKCanCor: Interjection
    "ECHO0": "PROPN",  # In one instance of "Big Echo"
    "F": "ADV",  # HKCanCor: Directional Locality
    "G": "X",  # HKCanCor: Morpheme
    "G1": "V",  # The first A in the "A-not-AB" pattern, where AB is a verb.
    "G2": "ADJ",  # The first A in "A-not-AB", where AB is an adjective.
    "H": "PROPN",  # HKCanCor: Prefix (aa3 阿 followed by a person name)
    "HILL0": "PROPN",  # In "Benny Hill"
    "I": "X",  # HKCanCor: Idiom
    "IG": "X",
    "J": "NOUN",  # HKCanCor: Abbreviation
    "JB": "ADJ",
    "JM": "NOUN",
    "JN": "NOUN",
    "JNS": "PROPN",
    "JNT": "PROPN",
    "JNZ": "PROPN",
    "K": "X",  # HKCanCor: Suffix (sing3 性 for nouns; dei6 地 for adverbs)
    "KONG": "PROPN",  # In "Hong Kong"
    "L": "X",  # Fixed Expression
    "L1": "X",
    "LG": "X",
    "M": "NUM",  # HKCanCor: Numeral
    "MG": "X",
    "MONTY0": "PROPN",  # In "Full Monty"
    "MOUNTAIN0": "PROPN",  # In "Blue Mountain"
    "N": "NOUN",  # Common Noun
    "N1": "DET",  # HKCanCor: only used for ne1 呢 determiner in the upstream data https://github.com/fcbond/hkcancor/blob/41f0631acda4f6a45460483ae23cad880edaacc8/data/utf8/FC-035_v2#L3455  # noqa: E501
    "NG": "NOUN",
    "NR": "PROPN",  # HKCanCor: Personal Name
    "NS": "PROPN",  # HKCanCor: Place Name
    "NSG": "PROPN",
    "NT": "PROPN",  # HKCanCor: Organization Name
    "NX": "NOUN",  # HKCanCor: Nominal Character String
    "NZ": "PROPN",  # HKCanCor: Other Proper Noun
    "O": "X",  # HKCanCor: Onomatopoeia
    "P": "ADP",  # HKCanCor: Preposition
    "PEPPER0": "PROPN",  # In "Chilli and Pepper"
    "Q": "NOUN",  # HKCanCor: Classifier
    "QG": "NOUN",  # HKCanCor: Classifier Morpheme
    "R": "PRON",  # HKCanCor: Pronoun
    "RG": "PRON",  # HKCanCor: Pronoun Morpheme
    "S": "NOUN",  # HKCanCor: Space Word
    "SOUND0": "PROPN",  # In "Manchester's Sound"
    "T": "ADV",  # HKCanCor: Time Word
    "TELECOM0": "PROPN",  # In "Hong Kong Telecom"
    "TG": "ADV",  # HKCanCor: Time Word Morpheme
    "TOUCH0": "PROPN",  # In "Don't Touch" (a magazine)
    "U": "PART",  # HKCanCor: Auxiliary (e.g., ge3 嘅 after an attributive adj)
    "UG": "PART",  # HKCanCor: Auxiliary Morpheme
    "U0": "PROPN",  # U as in "Hong Kong U" (= The University of Hong Kong)
    "V": "VERB",  # HKCanCor: Verb
    "V1": "VERB",
    "VD": "ADV",  # HKCanCor: Verb as Adverbial
    "VG": "VERB",
    "VK": "VERB",
    "VN": "NOUN",  # HKCanCor: Verb with Nominal Function
    "VU": "AUX",
    "VUG": "AUX",
    "W": "PUNCT",  # HKCanCor: Punctuation
    "X": "X",  # HKCanCor: Unclassified Item
    "XA": "ADJ",
    "XB": "ADJ",
    "XC": "CCONJ",
    "XD": "ADV",
    "XE": "INTJ",
    "XJ": "X",
    "XJB": "PROPN",
    "XJN": "NOUN",
    "XJNT": "PROPN",
    "XJNZ": "PROPN",
    "XJV": "VERB",
    "XJA": "ADJ",  # HKCanCor: Only for "A" (= additional) as in "A Maths" in the upstream data https://github.com/fcbond/hkcancor/blob/41f0631acda4f6a45460483ae23cad880edaacc8/data/utf8/FC-027_v2#L4712  # noqa: E501
    "XL1": "INTJ",
    "XM": "NUM",
    "XN": "NOUN",
    "XNG": "NOUN",
    "XNR": "PROPN",
    "XNS": "PROPN",
    "XNT": "PROPN",
    "XNX": "NOUN",
    "XNZ": "PROPN",
    "XO": "NOUN",  # HKCanCor: Only for Gik1lik1gu4 in the upstream data https://github.com/fcbond/hkcancor/blob/41f0631acda4f6a45460483ae23cad880edaacc8/data/utf8/FC-R006_v2#L881  # noqa: E501
    "XP": "ADP",
    "XQ": "NOUN",
    "XR": "PRON",
    "XS": "PROPN",
    "XT": "NOUN",
    "XV": "VERB",
    "XVG": "VERB",
    "XVN": "NOUN",
    "XX": "X",
    "Y": "PART",  # HKCanCor: Modal Particle
    "YG": "PART",  # HKCanCor: Modal Particle Morpheme
    "Y1": "PART",
    "Z": "ADJ",  # HKCanCor: Descriptive
}

# Add the Chinese full-length punctuation marks.
_MAP = {**_MAP, **{punct: "PUNCT" for punct in _PUNCTUATION_MARKS}}


def hkcancor_to_ud(tag: str = None):
    """Map a part-of-speech tag from HKCanCor to Universal Dependencies.

    HKCanCor uses a part-of-speech tagset of over 100 tags (46 of which
    are described at http://compling.hss.ntu.edu.sg/hkcancor/).
    For applications that would benefit from a less granular part-of-speech
    tagset (e.g., cross-linguistic natural language processing tasks),
    we can map the HKCanCor tagset to the Universal Dependencies v2 tagset
    with 17 tags (https://universaldependencies.org/u/pos/index.html)
    -- the purpose of this function.

    Any unrecognized tag is mapped to ``"X"``.

    .. versionadded:: 3.1.0

    Parameters
    ----------
    tag : str, optional
        A tag from the original HKCanCor annotated data.
        If not provided or ``None``, this function returns the entire
        dictionary of the tagset mapping from HKCanCor to UD.

    Returns
    -------
    str or dict[str, str]
        A tag from the Universal Dependencies v2 tagset, or a dictioary
        from HKCanCor to UD tags if no input is given.

    Examples
    --------
    >>> hkcancor_to_ud("V")
    'VERB'
    """
    if tag is None:
        return _MAP
    else:
        return _MAP.get(tag.strip().upper()) or "X"



import dataclasses
import functools
import os
from typing import List, Optional, Union

from pylangacq.chat import Reader, _params_in_docstring
from pylangacq.chat import read_chat as pylangacq_read_chat
from pylangacq.objects import Gra

from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.search import _perform_search
from pycantonese.util import _deprecate


_ENCODING = "utf-8"


@dataclasses.dataclass
class Token:
    """Token with attributes as parsed from a CHAT utterance.

    Attributes
    ----------
    word : str
        Word form of the token
    pos : str
        Part-of-speech tag
    jyutping : str
        Jyutping romanization
    mor : str
        Morphological information
    gra : Gra
        Grammatical relation
    """

    __slots__ = ("word", "pos", "jyutping", "mor", "gra")

    word: str
    pos: Optional[str]
    jyutping: Optional[str]
    mor: Optional[str]
    gra: Optional[Gra]


class CHATReader(Reader):
    """A reader for Cantonese CHAT corpus files.

    .. note:: Some of the methods are inherited from the parent class
        :class:`~pylangacq.Reader` for language acquisition,
        which may or may not be applicable to your use case.
    """

    @staticmethod
    def _preprocess_token(t) -> Token:
        # Examples from the CHILDES LeeWongLeung corpus, child mhz
        # e.g., mor is suk1&DIM=uncle, word is 叔叔
        # e.g., mor is ngo5-PL=I, word i 我

        try:
            jyutping_mor, _, eng = t.mor.partition("=")
        except AttributeError:
            return Token(t.word, t.pos, None, t.mor, t.gra)

        if "-" in jyutping_mor:
            jyutping, _, mor = jyutping_mor.partition("-")
        elif "&" in jyutping_mor:
            jyutping, _, mor = jyutping_mor.partition("&")
        else:
            jyutping = jyutping_mor
            mor = ""

        mor = f"{mor}={eng}" if eng else mor

        try:
            parse_jyutping(jyutping)
        except ValueError:
            jyutping = None

        return Token(t.word, t.pos, jyutping or None, mor or None, t.gra)

    @_params_in_docstring("participants", "exclude", "by_utterances", "by_files")
    def jyutping(
        self, participants=None, exclude=None, by_utterances=False, by_files=False
    ) -> Union[List[str], List[List[str]], List[List[List[str]]]]:
        """Return the data in Jyutping romanization.

        Parameters
        ----------

        Returns
        -------
        List[List[List[str]]] if both by_utterances and by_files are True
        List[List[str]] if by_utterances is True and by_files is False
        List[List[str]] if by_utterances is False and by_files is True
        List[str] if both by_utterances and by_files are False
        """
        tagged_sents = self.tokens(
            participants=participants,
            exclude=exclude,
            by_utterances=True,
            by_files=True,
        )
        result = [
            [
                [tagged_word.jyutping for tagged_word in tagged_sent]
                for tagged_sent in tagged_sents_for_file
            ]
            for tagged_sents_for_file in tagged_sents
        ]
        if by_files and by_utterances:
            pass
        elif by_files and not by_utterances:
            result = [self._flatten(list, f) for f in result]
        elif not by_files and by_utterances:
            result = self._flatten(list, result)
        else:
            # not by_files and not by_utterances
            result = self._flatten(list, (self._flatten(list, f) for f in result))
        return result

    def jyutping_sents(self, participants=None, exclude=None, by_files=False):
        _deprecate(
            "jyutping_sents", "jyutping with by_utterances=True", "3.2.0", "4.0.0"
        )
        return self.jyutping(
            participants=participants,
            exclude=exclude,
            by_utterances=True,
            by_files=by_files,
        )

    def jyutpings(
        self, participants=None, exclude=None, by_utterances=False, by_files=False
    ):
        _deprecate("jyutpings", "jyutping", "3.2.0", "4.0.0")
        return self.jyutping(
            participants=participants,
            exclude=exclude,
            by_utterances=by_utterances,
            by_files=by_files,
        )

    @staticmethod
    def _get_chars_from_sent(sent: List[str]) -> List[str]:
        result = []
        for word in sent:
            if word and "\u4e00" <= word[0] <= "\u9fff":
                result.extend(list(word))
            else:
                result.append(word)
        return result

    @_params_in_docstring("participants", "exclude", "by_utterances", "by_files")
    def characters(
        self, participants=None, exclude=None, by_utterances=False, by_files=False
    ) -> Union[List[str], List[List[str]], List[List[List[str]]]]:
        """Return the data in individual Chinese characters.

        Parameters
        ----------

        Returns
        -------
        List[List[List[str]]] if both by_utterances and by_files are True
        List[List[str]] if by_utterances is True and by_files is False
        List[List[str]] if by_utterances is False and by_files is True
        List[str] if both by_utterances and by_files are False
        """
        sents = self.words(
            participants=participants,
            exclude=exclude,
            by_utterances=True,
            by_files=True,
        )
        result = [
            [self._get_chars_from_sent(sent) for sent in sents_for_file]
            for sents_for_file in sents
        ]
        if by_files and by_utterances:
            pass
        elif by_files and not by_utterances:
            result = [self._flatten(list, f) for f in result]
        elif not by_files and by_utterances:
            result = self._flatten(list, result)
        else:
            # not by_files and not by_utterances
            result = self._flatten(list, (self._flatten(list, f) for f in result))
        return result

    def character_sents(self, participants=None, exclude=None, by_files=False):
        _deprecate(
            "character_sents", "characters with by_utterances=True", "3.2.0", "4.0.0"
        )
        return self.characters(
            participants=participants,
            exclude=exclude,
            by_utterances=True,
            by_files=by_files,
        )

    def search(
        self,
        *,
        onset=None,
        nucleus=None,
        coda=None,
        tone=None,
        initial=None,
        final=None,
        jyutping=None,
        character=None,
        pos=None,
        word_range=(0, 0),
        utterance_range=(0, 0),
        sent_range=(0, 0),  # Deprecated
        by_tokens=True,
        by_utterances=False,
        tagged=None,  # Deprecated
        sents=None,  # Deprecated
        participants=None,
        exclude=None,
        by_files=False,
    ):
        """Search the data for the given criteria.

        For examples, please see https://pycantonese.org/searches.html.

        Parameters
        ----------
        onset : str, optional
            Onset to search for. A regex is supported.
        nucleus : str, optional
            Nucleus to search for. A regex is supported.
        coda : str, optional
            Coda to search for. A regex is supported.
        tone : str, optional
            Tone to search for. A regex is supported.
        initial : str, optional
            Initial to search for. A regex is supported.
            An initial, a term more prevalent in traditional Chinese
            phonology, is the equivalent of an onset.
        final : str, optional
            Final to search for.
            A final, a term more prevalent in traditional Chinese
            phonology, is the equivalent of a nucleus plus a coda.
        jyutping : str, optional
            Jyutping romanization of one Cantonese character to search for.
            If the romanization contains more than one character, a ValueError
            is raised.
        character : str, optional
            One or more Cantonese characters (within a segmented word) to
            search for.
        pos : str, optional
            A part-of-speech tag to search for. A regex is supported.
        word_range : tuple[int, int], optional
            Span of words to the left and right of a matching word to include
            in the output. The default is `(0, 0)` to disable a range.
            If `sent_range` is used, `word_range` is ignored.
        utterance_range : Tuple[int, int], optional
            Span of utterances before and after an utterance containing a matching
            word to include in the output.
            If set to ``(0, 0)`` (the default), no utterance range output is generated.
            If `utterance_range` is used, `word_range` is ignored.
        sent_range : Tuple[int, int], optional
            [Deprecated; please use utterance_range instead]
        by_tokens : bool, optional
            If ``True`` (the default), words in the output are in the token form
            (i.e., with Jyutping and part-of-speech tags).
            Otherwise just words as text strings are returned.
        by_utterances : bool, optional
            If ``True`` (default is False), utterances containing matching words
            are returned. Otherwise, only matching words are returned.
        tagged : bool, optional
            [Deprecated; please use by_tokens instead]
        sents : bool, optional
            [Deprecated; please use by_utterances instead]
        participants : str or iterable[str], optional
            One or more participants to include in the search.
            If unspecified, all participants are included.
        exclude : str or iterable[str], optional
            One or more participants to exclude in the search.
            If unspecified, no participants are excluded.
        by_files : bool, optional
            If True (default: False), return data organized by the
            individual file paths.

        Returns
        -------
        list
        """
        if sent_range != (0, 0):
            _deprecate("sent_range", "utterance_range", "3.2.0", "4.0.0")
            if utterance_range != (0, 0):
                raise TypeError(
                    "Do not use both utterance_range and sent_range "
                    f"(you've passed in {utterance_range} and {sent_range}, "
                    f"respectively). "
                    f"Please use utterance_range; "
                    f"sent_range has been deprecated."
                )
            utterance_range = sent_range
        if tagged is not None:
            _deprecate("tagged", "by_tokens", "3.2.0", "4.0.0")
            by_tokens = tagged
        if sents is not None:
            _deprecate("sents", "by_utterances", "3.2.0", "4.0.0")
            by_utterances = sents

        tagged_sents = self.tokens(
            participants=participants,
            exclude=exclude,
            by_utterances=True,
            by_files=True,
        )
        result_by_files = _perform_search(
            tagged_sents,
            onset=onset,
            nucleus=nucleus,
            coda=coda,
            tone=tone,
            initial=initial,
            final=final,
            jyutping=jyutping,
            character=character,
            pos=pos,
            word_range=word_range,
            utterance_range=utterance_range,
            by_tokens=by_tokens,
            by_utterances=by_utterances,
        )

        if by_files:
            return result_by_files
        else:
            return self._flatten(list, result_by_files)


class _HKCanCorReader(CHATReader):
    """Corpus reader for HKCanCor specifically.

    We enforce uppercase for part-of-speech tags,
    because the original HKCanCor's tags have a mix of upper- and lowercase, e.g.,
    v and Vg, which makes it harder to perform a corpus search
    with a clean(er) regex.
    """

    @staticmethod
    def _preprocess_pos(pos: str) -> str:
        """Override the parent Reader class's method."""
        try:
            return pos.upper()
        except AttributeError:
            return pos


@functools.lru_cache(maxsize=1)
def hkcancor() -> CHATReader:
    """Create a corpus object for the Hong Kong Cantonese Corpus.

    Returns
    -------
    :class:`~pycantonese.CHATReader`
    """
    data_dir = os.path.join(os.path.dirname(__file__), "data", "hkcancor")
    reader = _HKCanCorReader.from_dir(data_dir)
    for f in reader._files:
        f.file_path = f.file_path.replace(data_dir, "").lstrip(os.sep)
    return reader


@_params_in_docstring("match", "exclude", "encoding", class_method=False)
def read_chat(
    path: str, match: str = None, exclude: str = None, encoding: str = _ENCODING
) -> CHATReader:
    """Read Cantonese CHAT data files.

    Parameters
    ----------
    path : str
        A path that points to one of the following:

        - ZIP file. Either a local ``.zip`` file path or a URL (one that begins with
          ``"https://"`` or ``"http://"``).
          URL example: ``"https://childes.talkbank.org/data/Biling/YipMatthews.zip"``
        - A local directory, for files under this directory recursively.
        - A single ``.cha`` CHAT file.

    Returns
    -------
    :class:`~pycantonese.CHATReader`
    """
    return pylangacq_read_chat(
        path, match=match, exclude=exclude, encoding=encoding, cls=CHATReader
    )



"""Tests related to the documentation."""


import pytest
import requests


@pytest.mark.parametrize(
    "url",
    [
        "https://talkbank.org/manuals/CHAT.pdf",
        "https://pylangacq.org/",
        "http://compling.hss.ntu.edu.sg/hkcancor/",
        "https://github.com/jacksonllee/pycantonese/blob/master/pycantonese/data/hkcancor/README.md",  # noqa: E501
        "https://childes.talkbank.org/data/Biling/YipMatthews.zip",
        "https://pylangacq.org/read.html",
        "https://pylangacq.org/headers.html",
        "https://docs.python.org/3/library/re.html",
        "https://www.lshk.org/jyutping",
        "https://www.ctan.org/pkg/tipa?lang=en",
        "https://pycantonese.org/papers/lee-chen-tsui-wicl3-slides-2016-03-12.pdf",
        "https://pycantonese.org/papers/lee-chen-tsui-wicl3-handout-2016-03-12.pdf",
        "https://pycantonese.org/papers/Lee-pycantonese-2015.html",
        "https://universaldependencies.org/u/pos/index.html",
        "https://pycantonese.org/index.html#links",
    ],
)
def test_urls_work(url):
    """URLs used in the documentation shouldn't be dead."""
    with requests.get(url) as r:
        assert r.status_code == 200



from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.util import _deprecate


ONSETS_TIPA = {
    "b": "p",
    "d": "t",
    "g": "k",
    "gw": "k\\super w ",
    "z": "ts",
    "p": "p\\super h ",
    "t": "t\\super h ",
    "k": "k\\super h ",
    "kw": "k\\super w\\super h ",
    "c": "ts\\super h ",
    "m": "m",
    "n": "n",
    "ng": "N",
    "f": "f",
    "h": "h",
    "s": "s",
    "l": "l",
    "w": "w",
    "j": "j",
    "": "",
}

FINALS_TIPA = {
    "i": "i",
    "ip": "ip\\textcorner ",
    "it": "it\\textcorner ",
    "ik": "Ik\\textcorner ",
    "im": "im",
    "in": "in",
    "ing": "IN",
    "iu": "iu",
    "yu": "y",
    "yut": "yt\\textcorner ",
    "yun": "yn",
    "u": "u",
    "ut": "ut\\textcorner ",
    "uk": "Uk\\textcorner ",
    "un": "un",
    "ung": "UN",
    "ui": "uY",
    "e": "E",
    "ek": "Ek\\textcorner ",
    "eng": "EN",
    "ei": "eI",
    "eot": "8t\\textcorner ",
    "eon": "8n",
    "eoi": "8Y",
    "oe": "\\oe ",
    "oek": "\\oe k\\textcorner ",
    "oeng": "\\oe N",
    "o": "O",
    "ot": "Ot\\textcorner ",
    "ok": "Ok\\textcorner ",
    "on": "On",
    "ong": "ON",
    "oi": "OY",
    "ou": "ou",
    "ap": "5p\\textcorner ",
    "at": "5t\\textcorner ",
    "ak": "5k\\textcorner ",
    "am": "5m",
    "an": "5n",
    "ang": "5N",
    "ai": "5I",
    "au": "5u",
    "aa": "a",
    "aap": "ap\\textcorner ",
    "aat": "at\\textcorner ",
    "aak": "ak\\textcorner ",
    "aam": "am",
    "aan": "an",
    "aang": "aN",
    "aai": "aI",
    "aau": "au",
    "m": "\\s{m}",
    "ng": "\\s{N}",
}

TONES_TIPA = {
    "1": "55",
    "2": "25",
    "3": "33",
    "4": "21",
    "5": "23",
    "6": "22",
}


def jyutping_to_tipa(jp_str):
    """Convert Jyutping romanization into LaTeX TIPA.

    .. versionadded:: 3.0.0
        This function replaces the deprecated equivalent ``jyutping2tipa``.

    Parameters
    ----------
    jp_str : str
        Jyutping romanization for one or multiple characters

    Returns
    -------
    list[str]

    Raises
    ------
    ValueError
        If the Jyutping romanization is illegal (e.g., with unrecognized
        elements).

    Examples
    --------
    >>> jyutping_to_tipa("gwong2dung1waa2")  # 廣東話, Cantonese  # doctest: +SKIP
    ['k\\super w ON25', 'tUN55', 'wa25']
    """  # noqa: E501
    jp_parsed_list = parse_jyutping(jp_str)
    tipa_list = []

    for jp_parsed in jp_parsed_list:
        # TODO: Separate "final" as "nucleus" and "coda" instead?
        tipa = ONSETS_TIPA[jp_parsed.onset] + FINALS_TIPA[jp_parsed.final]
        tipa = tipa.strip() + TONES_TIPA[jp_parsed.tone]
        tipa_list.append(tipa)

    return tipa_list


@_deprecate("jyutping2tipa", "jyutping_to_tipa", "3.0.0", "4.0.0")
def jyutping2tipa(*args, **kwargs):
    """Same as jyutping_to_tipa.

    .. deprecated:: 3.0.0
    """
    return jyutping_to_tipa(*args, **kwargs)



import pytest

from pycantonese import segment
from pycantonese.word_segmentation import Segmenter


@pytest.mark.parametrize(
    "chars, segmenter, expected",
    [
        ("廣東話容唔容易學？", None, ["廣東話", "容", "唔", "容易", "學", "？"]),
        (
            "廣東話容唔容易學？",
            Segmenter(allow={"容唔容易"}),
            ["廣東話", "容唔容易", "學", "？"],
        ),
        (
            "廣東話容唔容易學？",
            Segmenter(disallow={"廣東話"}),
            ["廣東", "話", "容", "唔", "容易", "學", "？"],
        ),
        (
            "廣東話容唔容易學？",
            Segmenter(max_word_length=2),
            ["廣東", "話", "容", "唔", "容易", "學", "？"],
        ),
        ("佢淨係識呃like", Segmenter(), ["佢", "淨係", "識", "呃like"]),
        ("個course超好grade", Segmenter(), ["個", "course", "超", "好", "grade"]),
        # Whitespace in input string has no effect.
        ("廣東 話容唔容 易學？ ", None, ["廣東話", "容", "唔", "容易", "學", "？"]),
        ("\n 廣東話容唔容易學？ ", None, ["廣東話", "容", "唔", "容易", "學", "？"]),
    ],
)
def test_segment(chars, segmenter, expected):
    actual = segment(chars, cls=segmenter)
    assert actual == expected



import os
import re
from setuptools import setup, find_packages


_THIS_DIR = os.path.dirname(__file__)

_VERSION = "3.2.0"


def _get_long_description():
    with open(os.path.join(_THIS_DIR, "README.rst"), encoding="utf8") as f:
        readme = f.read().strip()
    # PyPI / twine doesn't accept the `raw` directive in reStructuredText.
    long_description = re.sub(
        r"\.\. start-raw-directive[\s\S]+?\.\. end-raw-directive", "", readme
    )
    return long_description


def main():
    setup(
        name="pycantonese",
        version=_VERSION,
        description="PyCantonese: Cantonese Linguistics and NLP in Python",
        long_description=_get_long_description(),
        long_description_content_type="text/x-rst",
        url="https://pycantonese.org",
        project_urls={
            "Bug Tracker": "https://github.com/jacksonllee/pycantonese/issues",
            "Source Code": "https://github.com/jacksonllee/pycantonese",
        },
        download_url="https://pypi.org/project/pycantonese/#files",
        author="Jackson L. Lee",
        author_email="jacksonlunlee@gmail.com",
        license="MIT License",
        packages=find_packages(),
        keywords=[
            "computational linguistics",
            "natural language processing",
            "NLP",
            "Cantonese",
            "linguistics",
            "corpora",
            "speech",
            "language",
            "Chinese",
            "Jyutping",
        ],
        python_requires=">=3.6",
        setup_requires="setuptools>=39",
        install_requires=[
            "dataclasses ; python_version < '3.7'",
            "pylangacq==0.13.0",
            "wordseg==0.0.2",
        ],
        package_data={
            "pycantonese": [
                "data/hkcancor/*",
                "data/rime_cantonese/*",
                "pos_tagging/*.pickle",
            ],
        },
        data_files=[(".", ["README.rst", "LICENSE.txt", "CHANGELOG.md"])],
        zip_safe=False,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Information Technology",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: Chinese (Traditional)",
            "Natural Language :: Cantonese",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Human Machine Interfaces",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Text Processing",
            "Topic :: Text Processing :: Filters",
            "Topic :: Text Processing :: General",
            "Topic :: Text Processing :: Indexing",
            "Topic :: Text Processing :: Linguistic",
        ],
    )


if __name__ == "__main__":
    main()



import dataclasses
import re

from pycantonese.jyutping.parse_jyutping import parse_jyutping, _parse_final, Jyutping


def _jp_element_match(search_element, current_element):
    if search_element is None or re.search(search_element, current_element):
        return True
    else:
        return False


def _perform_search(
    tagged_sents,
    onset=None,
    nucleus=None,
    coda=None,
    tone=None,
    initial=None,
    final=None,
    jyutping=None,
    character=None,
    pos=None,
    word_range=(0, 0),
    utterance_range=(0, 0),
    by_tokens=True,
    by_utterances=False,
):
    """
    overall strategy: deal with jp (and all jp-related elements) first, and
                      then the character

    1. jp

    hierarchy of jp and associated search elements:
                      jp
                 /     |    \
      onset/initial  final  tone
                     /   \
                nucleus  coda

    lower search elements cannot be used together with dominating higher
    elements
    """
    # ensure tuple type: word_range and sent_range
    if not (type(word_range) == type(utterance_range) == tuple):
        raise ValueError("word_range and sent_range must be tuples")

    words_left, words_right = word_range
    sents_left, sents_right = utterance_range

    # ensure int type: words_left, words_right, sents_left, sents_right
    if not (
        type(words_left)
        == type(words_right)
        == type(sents_left)
        == type(sents_right)
        == int
    ):
        raise ValueError("int required for {words, sents}_{left, right}")

    if sents_left > 0 or sents_right > 0:
        by_utterances = True

    if tone is None:
        pass
    elif str(tone) not in "123456":
        raise ValueError(
            f"tone is used, but it's not one of {{1, 2, 3, 4, 5, 6}}: {tone}"
        )
    else:
        tone = str(tone)

    # determine what kinds of search we are doing
    character_search = False
    jp_search = False
    pos_search = False

    if character:
        character_search = True
    if onset or nucleus or coda or tone or final or jyutping:
        jp_search = True
    if pos:
        pos_search = True
    if not (character_search or jp_search or pos_search):
        raise ValueError("no search elements")

    # check if jyutping search is valid
    jp_search_tuple = (None, None, None, None)

    if jp_search:

        # ensure compatible jyutping search elements
        if final and (nucleus or coda):
            raise ValueError(
                "final cannot be used together with either nucleus or coda (or both)"
            )
        if jyutping and (onset or final or nucleus or coda or tone):
            raise ValueError(
                "jyutping cannot be used together with other Jyutping elements"
            )
        if (onset != initial) and onset and initial:
            raise ValueError("onset conflicts with initial")

        # onset/initial
        if initial:
            onset = initial

        # determine jp_search_tuple
        if jyutping:
            try:
                jp_search_list = parse_jyutping(jyutping)
            except ValueError:
                raise ValueError("invalid jyutping -- %s" % (repr(jyutping)))
            if len(jp_search_list) > 1:
                raise ValueError("only jyutping for one character is allowed")
            else:
                jp_search_tuple = jp_search_list[0]
        else:
            if final:
                nucleus, coda = _parse_final(final)
            jp_search_tuple = Jyutping(onset, nucleus, coda, tone)

    result = []

    for tagged_sents_for_file in tagged_sents:
        sent_word_index_pairs = []

        for i_sent, tagged_sent in enumerate(tagged_sents_for_file):

            for i_word, tagged_word in enumerate(tagged_sent):
                c_characters = tagged_word.word
                c_pos = tagged_word.pos
                c_jyutping = tagged_word.jyutping

                # determine character_search and pos_search
                if character_search:
                    character_match = character in c_characters
                else:
                    character_match = True

                # if 'V' in c_pos.upper():
                #     import pdb; pdb.set_trace()
                if pos_search:
                    pos_match = bool(re.search(pos, c_pos))
                else:
                    pos_match = True

                if not (character_match and pos_match):
                    continue

                # determine if jyutping matches c_jyutping
                jyutping_match = False

                if not jp_search:
                    jyutping_match = True
                elif not c_jyutping:
                    pass
                else:
                    try:
                        c_parsed_jyutpings = parse_jyutping(c_jyutping)
                    except ValueError:
                        continue

                    for c_parsed_jyutping in c_parsed_jyutpings:

                        booleans = [
                            _jp_element_match(search_, current_)
                            for search_, current_ in zip(
                                dataclasses.astuple(jp_search_tuple),
                                dataclasses.astuple(c_parsed_jyutping),
                            )
                        ]

                        if all(booleans):
                            jyutping_match = True
                        break

                if jyutping_match:
                    sent_word_index_pairs.append((i_sent, i_word))

        results_list = []

        for i_sent, i_word in sent_word_index_pairs:
            if not by_utterances:
                tagged_sent = tagged_sents_for_file[i_sent]
                i_word_start = i_word - words_left
                i_word_end = i_word + words_right + 1

                if i_word_start < 0:
                    i_word_start = 0
                if i_word_end > len(tagged_sent):
                    i_word_end = len(tagged_sent)

                words_wanted = tagged_sent[i_word_start:i_word_end]

                if not by_tokens:
                    words_wanted = [x[0] for x in words_wanted]

                if len(words_wanted) == 1:
                    words_wanted = words_wanted[0]

                results_list.append(words_wanted)
            else:
                i_sent_start = i_sent - sents_left
                i_sent_end = i_sent + sents_right + 1

                if i_sent_start < 0:
                    i_sent_start = 0
                if i_sent_end > len(tagged_sents_for_file):
                    i_sent_end = len(tagged_sents_for_file)

                sents_wanted = tagged_sents_for_file[i_sent_start:i_sent_end]

                if not by_tokens:
                    for i, sent in enumerate(sents_wanted[:]):
                        sents_wanted[i] = [x[0] for x in sent]

                if len(sents_wanted) == 1:
                    sents_wanted = sents_wanted[0]

                results_list.append(sents_wanted)

        result.append(results_list)

    return result



"""This script trains a part-of-speech tagger."""

import logging

from pycantonese import hkcancor
from pycantonese.pos_tagging import POSTagger
from pycantonese.pos_tagging.tagger import _PICKLE_PATH


_TAGGER_PARAMETERS = {
    "frequency_threshold": 10,
    "ambiguity_threshold": 0.9,
    "n_iter": 10,
}

# Several POS tags in HKCanCor are odd ones for proper nouns.
_FIX_HKCANCOR_TAGS = {
    "AIRWAYS0": "XNT",
    "AND": "XNZ",  # In "Chilli and Pepper"
    "BEAN0": "XNZ",  # In one instance of "Mr Bean"
    "CENTRE0": "XNT",  # In one instance of "career centre"
    "ECHO0": "XNT",  # In one instance of "Big Echo"
    "HILL0": "XNZ",  # In "Benny Hill"
    "KONG": "XJNT",  # In "Hong Kong"
    "MONTY0": "XN",  # In "Full Monty"
    "MOUNTAIN0": "XNZ",  # In "Blue Mountain"
    "PEPPER0": "XNZ",  # In "Chilli and Pepper"
    "SOUND0": "XNZ",  # In "Manchester's Sound"
    "TELECOM0": "XNT",  # In "Hong Kong Telecom"
    "TOUCH0": "XNZ",  # In "Don't Touch" (a magazine)
    "U0": "XNT",  # U as in "Hong Kong U" (= The University of Hong Kong)
}


def _get_tagged_sents():
    return [
        [(token.word, _FIX_HKCANCOR_TAGS.get(token.pos, token.pos)) for token in tokens]
        for tokens in hkcancor().tokens(by_utterances=True)
    ]


if __name__ == "__main__":
    logging.basicConfig(level="INFO")
    tagger = POSTagger(**_TAGGER_PARAMETERS)
    tagger.train(_get_tagged_sents(), save=_PICKLE_PATH)
